"use strict";(self["webpackChunkplatypush"]=self["webpackChunkplatypush"]||[]).push([[1706],{1706:function(l,a,s){s.r(a),s.d(a,{default:function(){return D}});var e=s(6252),n=s(3577),t=s(9963);const c=l=>((0,e.dD)("data-v-75353a60"),l=l(),(0,e.Cn)(),l),i={class:"entity ping-host-container"},o={class:"head"},d={class:"col-1 icon-container"},u=["textContent"],v={key:0,class:"body children attributes fade-in"},r={class:"child"},p=c((()=>(0,e._)("div",{class:"col-s-12 col-m-6 label"},[(0,e._)("div",{class:"name"},"Host")],-1))),C={class:"value"},_=["textContent"],h={key:0,class:"child"},m=c((()=>(0,e._)("div",{class:"col-s-12 col-m-6 label"},[(0,e._)("div",{class:"name"},"Ping")],-1))),f={class:"value"},k=["textContent"];function b(l,a,s,c,b,w){return(0,e.wg)(),(0,e.iD)("div",i,[(0,e._)("div",o,[(0,e._)("div",d,[(0,e._)("span",{class:(0,n.C_)(["icon",w.iconClass])},null,2)]),(0,e._)("div",{class:"col-10 name",onClick:a[0]||(a[0]=(0,t.iM)((l=>b.isCollapsed=!b.isCollapsed),["stop"]))},[(0,e._)("div",{class:"name",textContent:(0,n.zw)(l.value.name)},null,8,u)]),(0,e._)("div",{class:"col-1 collapse-toggler",onClick:a[1]||(a[1]=(0,t.iM)((l=>b.isCollapsed=!b.isCollapsed),["stop"]))},[(0,e._)("i",{class:(0,n.C_)(["fas",{"fa-chevron-down":b.isCollapsed,"fa-chevron-up":!b.isCollapsed}])},null,2)])]),b.isCollapsed?(0,e.kq)("",!0):((0,e.wg)(),(0,e.iD)("div",v,[(0,e._)("div",r,[p,(0,e._)("div",C,[(0,e._)("div",{class:"name",textContent:(0,n.zw)(l.value.name)},null,8,_)])]),l.value.reachable?((0,e.wg)(),(0,e.iD)("div",h,[m,(0,e._)("div",f,[(0,e._)("span",{textContent:(0,n.zw)(l.value.avg)},null,8,k),(0,e.Uk)(" ms ")])])):(0,e.kq)("",!0)]))])}var w=s(847),g={mixins:[w["default"]],data(){return{isCollapsed:!0}},computed:{iconClass(){return this.value.reachable?"reachable":"unreachable"}}},x=s(3744);const y=(0,x.Z)(g,[["render",b],["__scopeId","data-v-75353a60"]]);var D=y}}]);
//# sourceMappingURL=1706.1a5248ba.js.map