"use strict";(self["webpackChunkplatypush"]=self["webpackChunkplatypush"]||[]).push([[2853],{5169:function(t,e,a){a.d(e,{Z:function(){return y}});var n=a(6252),l=a(9963);const i=t=>((0,n.dD)("data-v-a248454a"),t=t(),(0,n.Cn)(),t),s={class:"tts-container"},u={class:"field text-container"},r=["disabled"],d={class:"field lang-container"},o=["disabled"],c={class:"field buttons"},p=["disabled"],g=i((()=>(0,n._)("i",{class:"fa fa-volume-up"},null,-1))),f=[g];function v(t,e,a,i,g,v){return(0,n.wg)(),(0,n.iD)("div",s,[(0,n._)("form",{onSubmit:e[0]||(e[0]=(0,l.iM)(((...t)=>v.talk&&v.talk(...t)),["prevent"]))},[(0,n._)("div",u,[(0,n._)("label",null,[(0,n._)("input",{type:"text",name:"text",placeholder:"Text to say",disabled:g.talking},null,8,r)])]),(0,n._)("div",d,[(0,n._)("label",null,[(0,n._)("input",{type:"text",name:"language",placeholder:"Language code",disabled:g.talking},null,8,o)])]),(0,n._)("div",c,[(0,n._)("button",{type:"submit",disabled:g.talking},f,8,p)])],32)])}var m=a(8637),b={name:"Panel",mixins:[m.Z],props:{pluginName:{type:String,required:!0}},data(){return{talking:!1}},methods:{async talk(t){const e=[...t.target.querySelectorAll("input")].reduce(((t,e)=>(e.value.length&&(t[e.name]=e.value),t)),{});this.talking=!0;try{await this.request(`${this.pluginName}.say`,e)}finally{this.talking=!1}}}},h=a(3744);const k=(0,h.Z)(b,[["render",v],["__scopeId","data-v-a248454a"]]);var y=k},7605:function(t,e,a){a.r(e),a.d(e,{default:function(){return d}});var n=a(6252);function l(t,e,a,l,i,s){const u=(0,n.up)("Panel");return(0,n.wg)(),(0,n.j4)(u,{"plugin-name":"tts.google"})}var i=a(5169),s={name:"Tts",components:{Panel:i.Z}},u=a(3744);const r=(0,u.Z)(s,[["render",l]]);var d=r}}]);
//# sourceMappingURL=2853.99de82eb.js.map