"use strict";(self["webpackChunkplatypush"]=self["webpackChunkplatypush"]||[]).push([[446],{446:function(e,t,i){i.r(t),i.d(t,{default:function(){return b}});var s=i(6252),n=i(3577);const a={class:"camera component-row"},r={class:"feed-container",ref:"container"},c=["textContent"],l=["src"],o={key:2},u=["src"],p={class:"controls"},d={key:0,class:"fa fa-play"},g={key:1,class:"fa fa-pause"};function f(e,t,i,f,v,m){return(0,s.wg)(),(0,s.iD)("div",a,[(0,s._)("div",r,[v.visible?(0,s.kq)("",!0):((0,s.wg)(),(0,s.iD)("div",{key:0,class:"no-content",textContent:(0,n.zw)(i.name)},null,8,c)),v.visible&&"image"===i.type?((0,s.wg)(),(0,s.iD)("img",{key:1,alt:"Camera feed",src:m.imgUrl},null,8,l)):v.visible&&"video"===i.type?((0,s.wg)(),(0,s.iD)("video",o,[(0,s._)("source",{src:i.src},null,8,u)])):(0,s.kq)("",!0)],512),(0,s._)("div",p,[(0,s._)("button",{class:"toggle-btn",onClick:t[0]||(t[0]=e=>v.visible=!v.visible)},[v.visible?((0,s.wg)(),(0,s.iD)("i",g)):((0,s.wg)(),(0,s.iD)("i",d))])])])}var v=i(8637),m={name:"Camera",mixins:[v.Z],props:{src:{type:String,required:!0},type:{type:String,default:"image"},name:{type:String}},computed:{imgUrl(){if("image"===this.type)return this.src+(this.src.indexOf("?")>0?"&":"?")+"_t="+(new Date).getTime().toString()}},data(){return{visible:!1}}},y=i(3744);const k=(0,y.Z)(m,[["render",f],["__scopeId","data-v-0fad5251"]]);var b=k}}]);
//# sourceMappingURL=446.cdd8637e.js.map