"use strict";(self["webpackChunkplatypush"]=self["webpackChunkplatypush"]||[]).push([[8784],{8784:function(n,e,t){t.r(e),t.d(e,{default:function(){return o}});var r=t(6252);function u(n,e,t,u,a,p){const s=(0,r.up)("Panel");return(0,r.wg)(),(0,r.j4)(s,{"plugin-name":"torrent"})}var a=t(6429),p={name:"Torrent",components:{Panel:a.Z}},s=t(3744);const c=(0,s.Z)(p,[["render",u]]);var o=c}}]);
//# sourceMappingURL=8784.5892bf2a.js.map