mock_imports = [
    "Adafruit_IO",
    "Adafruit_Python_DHT",
    "Leap",
    "PIL",
    "PyOBEX",
    "PyOBEX.client",
    "RPLCD",
    "RPi.GPIO",
    "TheengsDecoder",
    "aiofiles",
    "aiofiles.os",
    "aiohttp",
    "aioxmpp",
    "apiclient",
    "async_lru",
    "avs",
    "bcrypt",
    "bleak",
    "bluetooth",
    "bluetooth_numbers",
    "buienradar",
    "cpuinfo",
    "croniter",
    "cups",
    "cv2",
    "cwiid",
    "dbus",
    "deepspeech",
    "defusedxml",
    "dns",
    "docutils",
    "envirophat",
    "feedparser",
    "gevent.wsgi",
    "gi",
    "gi.repository",
    "google",
    "google.api_core",
    "google.auth",
    "google.assistant.embedded",
    "google.assistant.library",
    "google.assistant.library.event",
    "google.assistant.library.file_helpers",
    "google.cloud",
    "google.oauth2.credentials",
    "googlesamples",
    "googlesamples.assistant.grpc.audio_helpers",
    "gps",
    "graphyte",
    "grpc",
    "gunicorn",
    "hid",
    "httplib2",
    "icalendar",
    "imapclient",
    "inotify",
    "inputs",
    "irc",
    "irc.bot",
    "irc.client",
    "irc.connection",
    "irc.events",
    "irc.strings",
    "kafka",
    "keras",
    "linode_api4",
    "luma",
    "marshmallow_dataclass",
    "mpd",
    "ndef",
    "nfc",
    "nio",
    "numpy",
    "oauth2client",
    "oauth2client",
    "omxplayer",
    "openzwave",
    "paho",
    "pandas",
    "paramiko",
    "picamera",
    "plexapi",
    "pmw3901",
    "psutil",
    "pvcheetah",
    "pvleopard",
    "pvorca",
    "pvporcupine",
    "pvrhino",
    "pyHS100",
    "pyaudio",
    "pychromecast",
    "pyclip",
    "pydbus",
    "pydub",
    "pyfirmata2",
    "pyngrok",
    "pyotp",
    "pysmartthings",
    "pyzbar",
    "qrcode",
    "rtmidi",
    "samsungtvws",
    "serial",
    "simple_websocket",
    "smartcard",
    "sounddevice",
    "soundfile",
    "telegram",
    "telegram.ext",
    "tenacity",
    "tensorflow",
    "todoist",
    "trello",
    "twilio",
    "uvicorn",
    "watchdog",
    "wave",
    "websockets",
    "zeroconf",
]
"""
List of modules that should be mocked when building the documentation or running tests.
"""
