"use strict";(self["webpackChunkplatypush"]=self["webpackChunkplatypush"]||[]).push([[5329],{5329:function(e,l,a){a.r(l),a.d(l,{default:function(){return le}});var s=a(6252),t=a(3577),n=a(9963);const i=e=>((0,s.dD)("data-v-d7813182"),e=e(),(0,s.Cn)(),e),v={class:"icon"},c={class:"label"},d=["textContent"],u={class:"value-and-toggler"},o=["textContent"],r={key:0,class:"body children attributes fade-in"},_={key:0,class:"child"},C=i((()=>(0,s._)("div",{class:"label"},[(0,s._)("div",{class:"name"},"Total")],-1))),m={class:"value"},k=["textContent"],h={key:1,class:"child"},p=i((()=>(0,s._)("div",{class:"label"},[(0,s._)("div",{class:"name"},"Available")],-1))),w={class:"value"},x=["textContent"],f={key:2,class:"child"},y=i((()=>(0,s._)("div",{class:"label"},[(0,s._)("div",{class:"name"},"Used")],-1))),z={class:"value"},b=["textContent"],g={key:3,class:"child"},D=i((()=>(0,s._)("div",{class:"label"},[(0,s._)("div",{class:"name"},"Free")],-1))),S={class:"value"},q=["textContent"],I={key:4,class:"child"},M=i((()=>(0,s._)("div",{class:"label"},[(0,s._)("div",{class:"name"},"Active")],-1))),A={class:"value"},E=["textContent"],B={key:5,class:"child"},F=i((()=>(0,s._)("div",{class:"label"},[(0,s._)("div",{class:"name"},"Inactive")],-1))),T={class:"value"},U=["textContent"],W={key:6,class:"child"},Z=i((()=>(0,s._)("div",{class:"label"},[(0,s._)("div",{class:"name"},"Buffers")],-1))),j={class:"value"},G=["textContent"],H={key:7,class:"child"},J=i((()=>(0,s._)("div",{class:"label"},[(0,s._)("div",{class:"name"},"Cached")],-1))),K={class:"value"},L=["textContent"],N={key:8,class:"child"},O=i((()=>(0,s._)("div",{class:"label"},[(0,s._)("div",{class:"name"},"Shared")],-1))),P={class:"value"},Q=["textContent"];function R(e,l,a,i,R,V){const X=(0,s.up)("EntityIcon");return(0,s.wg)(),(0,s.iD)("div",{class:(0,t.C_)(["entity memory-stats-container",{expanded:!R.isCollapsed}])},[(0,s._)("div",{class:"head",onClick:l[1]||(l[1]=(0,n.iM)((e=>R.isCollapsed=!R.isCollapsed),["stop"]))},[(0,s._)("div",v,[(0,s.Wm)(X,{entity:e.value,loading:e.loading,error:e.error},null,8,["entity","loading","error"])]),(0,s._)("div",c,[(0,s._)("div",{class:"name",textContent:(0,t.zw)(e.value.name)},null,8,d)]),(0,s._)("div",u,[(0,s._)("div",{class:"value",textContent:(0,t.zw)(Math.round(100*e.value.percent,1)+"%")},null,8,o),(0,s._)("div",{class:"collapse-toggler",onClick:l[0]||(l[0]=(0,n.iM)((e=>R.isCollapsed=!R.isCollapsed),["stop"]))},[(0,s._)("i",{class:(0,t.C_)(["fas",{"fa-chevron-down":R.isCollapsed,"fa-chevron-up":!R.isCollapsed}])},null,2)])])]),R.isCollapsed?(0,s.kq)("",!0):((0,s.wg)(),(0,s.iD)("div",r,[null!=e.value.total?((0,s.wg)(),(0,s.iD)("div",_,[C,(0,s._)("div",m,[(0,s._)("div",{class:"name",textContent:(0,t.zw)(e.convertSize(e.value.total))},null,8,k)])])):(0,s.kq)("",!0),null!=e.value.available?((0,s.wg)(),(0,s.iD)("div",h,[p,(0,s._)("div",w,[(0,s._)("div",{class:"name",textContent:(0,t.zw)(e.convertSize(e.value.available))},null,8,x)])])):(0,s.kq)("",!0),null!=e.value.used?((0,s.wg)(),(0,s.iD)("div",f,[y,(0,s._)("div",z,[(0,s._)("div",{class:"name",textContent:(0,t.zw)(e.convertSize(e.value.used))},null,8,b)])])):(0,s.kq)("",!0),null!=e.value.free?((0,s.wg)(),(0,s.iD)("div",g,[D,(0,s._)("div",S,[(0,s._)("div",{class:"name",textContent:(0,t.zw)(e.convertSize(e.value.free))},null,8,q)])])):(0,s.kq)("",!0),null!=e.value.active?((0,s.wg)(),(0,s.iD)("div",I,[M,(0,s._)("div",A,[(0,s._)("div",{class:"name",textContent:(0,t.zw)(e.convertSize(e.value.active))},null,8,E)])])):(0,s.kq)("",!0),null!=e.value.inactive?((0,s.wg)(),(0,s.iD)("div",B,[F,(0,s._)("div",T,[(0,s._)("div",{class:"name",textContent:(0,t.zw)(e.convertSize(e.value.inactive))},null,8,U)])])):(0,s.kq)("",!0),null!=e.value.buffers?((0,s.wg)(),(0,s.iD)("div",W,[Z,(0,s._)("div",j,[(0,s._)("div",{class:"name",textContent:(0,t.zw)(e.convertSize(e.value.buffers))},null,8,G)])])):(0,s.kq)("",!0),null!=e.value.cached?((0,s.wg)(),(0,s.iD)("div",H,[J,(0,s._)("div",K,[(0,s._)("div",{class:"name",textContent:(0,t.zw)(e.convertSize(e.value.cached))},null,8,L)])])):(0,s.kq)("",!0),null!=e.value.shared?((0,s.wg)(),(0,s.iD)("div",N,[O,(0,s._)("div",P,[(0,s._)("div",{class:"name",textContent:(0,t.zw)(e.convertSize(e.value.shared))},null,8,Q)])])):(0,s.kq)("",!0)]))],2)}var V=a(847),X=a(4967),Y={name:"MemoryStats",components:{EntityIcon:X["default"]},mixins:[V["default"]],data(){return{isCollapsed:!0}}},$=a(3744);const ee=(0,$.Z)(Y,[["render",R],["__scopeId","data-v-d7813182"]]);var le=ee}}]);
//# sourceMappingURL=5329.1dc4e553.js.map