"use strict";(self["webpackChunkplatypush"]=self["webpackChunkplatypush"]||[]).push([[8769],{8769:function(n,e,t){t.r(e),t.d(e,{default:function(){return h}});var a=t(6252),i=t(3577);const c={class:"entity device-container"},l={class:"head"},s={class:"icon"},o={class:"label"},r=["textContent"];function u(n,e,t,u,d,v){const p=(0,a.up)("EntityIcon");return(0,a.wg)(),(0,a.iD)("div",c,[(0,a._)("div",l,[(0,a._)("div",s,[(0,a.Wm)(p,{entity:n.value,loading:n.loading,error:n.error},null,8,["entity","loading","error"])]),(0,a._)("div",o,[(0,a._)("div",{class:"name",textContent:(0,i.zw)(n.value.name)},null,8,r)])])])}var d=t(847),v=t(4967),p={name:"Device",components:{EntityIcon:v["default"]},mixins:[d["default"]]},f=t(3744);const y=(0,f.Z)(p,[["render",u],["__scopeId","data-v-07323f6c"]]);var h=y}}]);
//# sourceMappingURL=8769.52fe04f5.js.map