"use strict";(self["webpackChunkplatypush"]=self["webpackChunkplatypush"]||[]).push([[169],{169:function(t,e,n){n.r(e),n.d(e,{default:function(){return x}});var l=n(6252),a=n(3577);const i={class:"entity sensor-container"},s={class:"head"},u={class:"icon"},o={class:"label"},r=["textContent"],c=["textContent"];function d(t,e,n,d,v,p){const f=(0,l.up)("EntityIcon");return(0,l.wg)(),(0,l.iD)("div",i,[(0,l._)("div",s,[(0,l._)("div",u,[(0,l.Wm)(f,{entity:t.value,loading:t.loading,error:t.error},null,8,["entity","loading","error"])]),(0,l._)("div",o,[(0,l._)("div",{class:"name",textContent:(0,a.zw)(t.value.name)},null,8,r)]),(0,l._)("div",{class:"value-container",textContent:(0,a.zw)(p.displayValue)},null,8,c)])])}var v=n(847),p=n(4967),f={name:"PercentSensor",components:{EntityIcon:p["default"]},mixins:[v["default"]],computed:{displayValue(){if(null==this.value.value)return null;let t=100*this.value.value;return(t.toString()==t.toFixed(0)?t.toFixed(0):t.toFixed(1))+"%"}}},y=n(3744);const h=(0,y.Z)(f,[["render",d],["__scopeId","data-v-1b6c81c2"]]);var x=h}}]);
//# sourceMappingURL=169.92c1a438.js.map