"""Base alembic version

Revision ID: d030953a871d
Revises:
Create Date: 2023-04-28 22:32:49.460118

"""


# revision identifiers, used by Alembic.
revision = 'd030953a871d'
down_revision = None
branch_labels = None
depends_on = None


def upgrade() -> None:
    pass


def downgrade() -> None:
    pass
