"use strict";(self["webpackChunkplatypush"]=self["webpackChunkplatypush"]||[]).push([[4221],{4221:function(n,t,c){c.r(t),c.d(t,{default:function(){return p}});var e=c(6252);const u=n=>((0,e.dD)("data-v-7c29571d"),n=n(),(0,e.Cn)(),n),o={class:"row plugin execute-container"},a=u((()=>(0,e._)("h1",null,"Execute Action",-1)));function r(n,t,c,u,r,i){const s=(0,e.up)("ActionEditor");return(0,e.wg)(),(0,e.iD)("div",o,[(0,e._)("main",null,[a,(0,e.Wm)(s)])])}var i=c(1807),s={components:{ActionEditor:i.Z}},d=c(3744);const l=(0,d.Z)(s,[["render",r],["__scopeId","data-v-7c29571d"]]);var p=l}}]);
//# sourceMappingURL=4221.b910d558.js.map