"use strict";(self["webpackChunkplatypush"]=self["webpackChunkplatypush"]||[]).push([[2892,6640],{2892:function(a,e,n){n.r(e),n.d(e,{default:function(){return p}});var t=n(6252);const r={class:"media-file-browser"};function i(a,e,n,i,s,o){const u=(0,t.up)("Loading"),c=(0,t.up)("Browser");return(0,t.wg)(),(0,t.iD)("div",r,[a.loading?((0,t.wg)(),(0,t.j4)(u,{key:0})):((0,t.wg)(),(0,t.j4)(c,{key:1,"is-media":!0,filter:a.filter,"has-back":!0,onBack:e[0]||(e[0]=e=>a.$emit("back")),onPathChange:e[1]||(e[1]=e=>a.$emit("path-change",e)),onPlay:e[2]||(e[2]=e=>a.$emit("play",e))},null,8,["filter"]))])}var s=n(2381),o=n(6791),u=n(6640),c={mixins:[u["default"]],components:{Browser:s.Z,Loading:o.Z}},l=n(3744);const d=(0,l.Z)(c,[["render",i],["__scopeId","data-v-5f4a36da"]]);var p=d},6640:function(a,e,n){n.r(e),n.d(e,{default:function(){return s}});var t=n(8637),r={emits:["back","path-change","play"],mixins:[t.Z],props:{filter:{type:String,default:""}},data(){return{loading:!1}}};const i=r;var s=i}}]);
//# sourceMappingURL=2892.d56bb8af.js.map