"use strict";(self["webpackChunkplatypush"]=self["webpackChunkplatypush"]||[]).push([[2460],{2460:function(n,t,e){e.r(t),e.d(t,{default:function(){return C}});var a=e(6252),l=e(3577);const c={class:"entity cpu-container"},s={class:"head"},o={class:"col-1 icon"},u={class:"label"},i=["textContent"],r={class:"value-container"},d=["textContent"];function v(n,t,e,v,p,f){const _=(0,a.up)("EntityIcon");return(0,a.wg)(),(0,a.iD)("div",c,[(0,a._)("div",s,[(0,a._)("div",o,[(0,a.Wm)(_,{entity:n.value,loading:n.loading,error:n.error},null,8,["entity","loading","error"])]),(0,a._)("div",u,[(0,a._)("div",{class:"name",textContent:(0,l.zw)(n.value.name)},null,8,i)]),(0,a._)("div",r,[(0,a._)("div",{class:"value",textContent:(0,l.zw)(Math.round(100*n.value.percent,1)+"%")},null,8,d)])])])}var p=e(847),f=e(4967),_={name:"Cpu",components:{EntityIcon:f["default"]},mixins:[p["default"]]},h=e(3744);const y=(0,h.Z)(_,[["render",v],["__scopeId","data-v-d3cf6cca"]]);var C=y}}]);
//# sourceMappingURL=2460.2a8400ba.js.map