"use strict";(self["webpackChunkplatypush"]=self["webpackChunkplatypush"]||[]).push([[6362],{6362:function(t,e,n){n.r(e),n.d(e,{default:function(){return w}});var l=n(6252),a=n(3577);const u={class:"entity sensor-container"},s={class:"head"},i={class:"icon"},o={class:"label"},c=["textContent"],r={key:0,class:"value-container"},v=["textContent"],d=["textContent"];function p(t,e,n,p,m,h){const y=(0,l.up)("EntityIcon");return(0,l.wg)(),(0,l.iD)("div",u,[(0,l._)("div",s,[(0,l._)("div",i,[(0,l.Wm)(y,{entity:t.value,loading:t.loading,error:t.error},null,8,["entity","loading","error"])]),(0,l._)("div",o,[(0,l._)("div",{class:"name",textContent:(0,a.zw)(t.value.name)},null,8,c)]),null!=h.computedValue?((0,l.wg)(),(0,l.iD)("div",r,[(0,l._)("span",{class:"value",textContent:(0,a.zw)(h.computedValue)},null,8,v),null!=t.value.unit?((0,l.wg)(),(0,l.iD)("span",{key:0,class:"unit",textContent:(0,a.zw)(t.value.unit)},null,8,d)):(0,l.kq)("",!0)])):(0,l.kq)("",!0)])])}var m=n(847),h=n(4967),y={name:"Sensor",components:{EntityIcon:h["default"]},mixins:[m["default"]],computed:{computedValue(){return null!=this.value.value?this.value.value:this.value._value}}},f=n(3744);const k=(0,f.Z)(y,[["render",p],["__scopeId","data-v-3b38610c"]]);var w=k}}]);
//# sourceMappingURL=6362.4870392f.js.map