"use strict";(self["webpackChunkplatypush"]=self["webpackChunkplatypush"]||[]).push([[3559],{3559:function(e,n,t){t.r(n),t.d(n,{default:function(){return k}});var l=t(6252),a=t(3577);const u={class:"entity link-quality-container"},i={class:"head"},s={class:"icon"},r={class:"label"},c=["textContent"],o={class:"value-container"},v=["textContent"];function d(e,n,t,d,p,f){const h=(0,l.up)("EntityIcon");return(0,l.wg)(),(0,l.iD)("div",u,[(0,l._)("div",i,[(0,l._)("div",s,[(0,l.Wm)(h,{entity:e.value,loading:e.loading,error:e.error},null,8,["entity","loading","error"])]),(0,l._)("div",r,[(0,l._)("div",{class:"name",textContent:(0,a.zw)(e.value.name)},null,8,c)]),(0,l._)("div",o,[null!=f.valuePercent?((0,l.wg)(),(0,l.iD)("span",{key:0,class:"value",textContent:(0,a.zw)(f.valuePercent+"%")},null,8,v)):(0,l.kq)("",!0)])])])}var p=t(847),f=t(4967),h={name:"LinkQuality",components:{EntityIcon:f["default"]},mixins:[p["default"]],computed:{valuePercent(){if(null==this.value?.value)return null;const e=this.value.min||0,n=this.value.max||100;return(100*this.value.value/(n-e)).toFixed(0)}}},y=t(3744);const m=(0,y.Z)(h,[["render",d],["__scopeId","data-v-66f207d9"]]);var k=m}}]);
//# sourceMappingURL=3559.61ebd182.js.map