"use strict";(self["webpackChunkplatypush"]=self["webpackChunkplatypush"]||[]).push([[1088],{1088:function(n,e,t){t.r(e),t.d(e,{default:function(){return o}});var u=t(6252);function a(n,e,t,a,r,p){const s=(0,u.up)("Zwave");return(0,u.wg)(),(0,u.j4)(s,{"plugin-name":"zwave.mqtt"})}var r=t(7880),p={components:{Zwave:r.Z}},s=t(3744);const c=(0,s.Z)(p,[["render",a]]);var o=c}}]);
//# sourceMappingURL=1088.f184400f.js.map