"use strict";(self["webpackChunkplatypush"]=self["webpackChunkplatypush"]||[]).push([[2200],{2200:function(s,i,e){e.r(i),e.d(i,{default:function(){return b}});var c=e(6252);const a={class:"youtube-views-browser grid"},l=(0,c._)("div",{class:"icon"},[(0,c._)("i",{class:"fas fa-rss"})],-1),t=(0,c._)("div",{class:"name"},"Feed",-1),n=[l,t],r=(0,c._)("div",{class:"icon"},[(0,c._)("i",{class:"fas fa-list"})],-1),u=(0,c._)("div",{class:"name"},"Playlists",-1),d=[r,u],o=(0,c._)("div",{class:"icon"},[(0,c._)("i",{class:"fas fa-user"})],-1),v=(0,c._)("div",{class:"name"},"Subscriptions",-1),f=[o,v];function _(s,i,e,l,t,r){return(0,c.wg)(),(0,c.iD)("div",a,[(0,c._)("div",{class:"item",onClick:i[0]||(i[0]=i=>s.$emit("select","feed"))},n),(0,c._)("div",{class:"item",onClick:i[1]||(i[1]=i=>s.$emit("select","playlists"))},d),(0,c._)("div",{class:"item",onClick:i[2]||(i[2]=i=>s.$emit("select","subscriptions"))},f)])}var m={emits:["select"]},p=e(3744);const k=(0,p.Z)(m,[["render",_]]);var b=k}}]);
//# sourceMappingURL=2200.a7bc2894.js.map