"use strict";(self["webpackChunkplatypush"]=self["webpackChunkplatypush"]||[]).push([[3322],{3322:function(a,e,n){n.r(e),n.d(e,{default:function(){return p}});var c=n(6252);const s={class:"entity weather-icon-container"},t=["src","alt"],r=["src","alt"];function l(a,e,n,l,u,i){return(0,c.wg)(),(0,c.iD)("span",s,[n.value.icon?((0,c.wg)(),(0,c.iD)("img",{key:0,src:`/icons/openweathermap/dark/${n.value.icon}.png`,alt:n.value?.summary,class:"weather-icon"},null,8,t)):n.value.image?((0,c.wg)(),(0,c.iD)("img",{key:1,src:n.value.image,alt:n.value?.summary,class:"weather-icon"},null,8,r)):(0,c.kq)("",!0)])}var u={props:{value:Object}},i=n(3744);const o=(0,i.Z)(u,[["render",l],["__scopeId","data-v-1c0bfb77"]]);var p=o}}]);
//# sourceMappingURL=3322.dadc23cc.js.map