"use strict";(self["webpackChunkplatypush"]=self["webpackChunkplatypush"]||[]).push([[2614],{2614:function(e,n,a){a.r(n),a.d(n,{default:function(){return i}});var u=a(6252);function t(e,n,a,t,c,r){const p=(0,u.up)("Media");return(0,u.wg)(),(0,u.j4)(p,{"plugin-name":"media.vlc"})}var c=a(8913),r={name:"MediaVlc",components:{Media:c["default"]}},p=a(3744);const s=(0,p.Z)(r,[["render",t]]);var i=s}}]);
//# sourceMappingURL=2614.b0dbc28c.js.map