"use strict";(self["webpackChunkplatypush"]=self["webpackChunkplatypush"]||[]).push([[2217],{2217:function(n,t,e){e.r(t),e.d(t,{default:function(){return y}});var a=e(6252),l=e(3577);const s={class:"entity cpu-times-container"},i={class:"head"},c={class:"col-1 icon"},o={class:"col-11 label"},r=["textContent"];function u(n,t,e,u,d,p){const v=(0,a.up)("EntityIcon");return(0,a.wg)(),(0,a.iD)("div",s,[(0,a._)("div",i,[(0,a._)("div",c,[(0,a.Wm)(v,{entity:n.value,loading:n.loading,error:n.error},null,8,["entity","loading","error"])]),(0,a._)("div",o,[(0,a._)("div",{class:"name",textContent:(0,l.zw)(n.value.name)},null,8,r)])])])}var d=e(847),p=e(4967),v={name:"CpuTimes",components:{EntityIcon:p["default"]},mixins:[d["default"]]},f=e(3744);const m=(0,f.Z)(v,[["render",u],["__scopeId","data-v-4667e342"]]);var y=m}}]);
//# sourceMappingURL=2217.d60fbb58.js.map