"use strict";(self["webpackChunkplatypush"]=self["webpackChunkplatypush"]||[]).push([[9091],{9091:function(t,e,o){o.r(e),o.d(e,{default:function(){return l}});var a=o(6252);const i={class:"no-token"},d=(0,a.uE)('<div class="title" data-v-42457341> No <code data-v-42457341>auth_token</code> found in the YouTube configuration. </div><div class="description" data-v-42457341> This integration requires an <code data-v-42457341>auth_token</code> to be set in the <code data-v-42457341>youtube</code> section of the configuration file in order to access your playlists and subscriptions.<br data-v-42457341><br data-v-42457341> Piped auth tokens are currently supported. You can retrieve one through the following procedure: <ol data-v-42457341><li data-v-42457341>Login to your configured Piped instance.</li><li data-v-42457341>Copy the RSS/Atom feed URL on the <i data-v-42457341>Feed</i> tab.</li><li data-v-42457341>Copy the <code data-v-42457341>auth_token</code> query parameter from the URL.</li><li data-v-42457341> Enter it in the <code data-v-42457341>auth_token</code> field in the <code data-v-42457341>youtube</code> section of the configuration file. </li></ol></div>',2),n=[d];function r(t,e){return(0,a.wg)(),(0,a.iD)("div",i,n)}var u=o(3744);const c={},s=(0,u.Z)(c,[["render",r],["__scopeId","data-v-42457341"]]);var l=s}}]);
//# sourceMappingURL=9091.b9371c68.js.map